         'PCBook3.BAS
         'Utility to print ASCII text files to LaserJet Series II, IIp or III
         '   in booklet format
         '
         'Copyright 1990 PC Magazine - Ziff Davis - Jay Munro
         'Written by Jay Munro
         'Modifications by Mike Shaffer, Savant Technologies, Inc.
         '        - Allows use of HP's duplexing printers (IID and IIID) for
         '          automatic page flipping...
         '
         '===========================================================================
         'LaserJet programming concepts employed:
         '       Setting orientation & font style
         '       Locating LaserJet cursor
         '       LaserJet Macro setup and use
         '===========================================================================
         'General programming concepts
         '       Building index arrays
         '       Using files for printing
         '
         'Compiler syntax:
         '    BC /o /S PCBook3.BAS;
         '    Link /NOD/NOE PCBook3,,,pdq;
         '
         '$include: 'pdqdecl.bas'
         DEFINT A-Z
         DECLARE SUB BuildArray (PtrArray&(), Pgcount%)
         DECLARE SUB EndMacro (Num$)                    'End of macro commands
         DECLARE SUB Header (Page%)                     'Print Header
         DECLARE SUB LJLocate (X%, Y%)                  'Laserjet cursor locate
         DECLARE SUB PrintSetup ()                      'Set up macros, fonts
         DECLARE SUB PrintLogo ()                       'Credits
         DECLARE SUB StartMacro (Num$)                  'Start of macro commands

         TYPE Flags                                     'Misc flag variables
           CurDate   AS INTEGER
           DoHeader  AS INTEGER
           FileTitle AS INTEGER
           LineLen   AS INTEGER
           LineWrap  AS INTEGER
           PgNumber  AS INTEGER
           Dplx      AS INTEGER
         END TYPE

         'Share variables with subs

         DIM SHARED ESC$, ESCANDF$, FF$, LF$, FileName$,csr%
         DIM SHARED PC AS Flags

         REDIM PtrArray&(513)                           'total number of pages (512)


         '============== Set some constant variables
         ESC$ = CHR$(27)                                'Standard ESC code
         ESCANDF$ = CHR$(27) + "&f"
         FF$ = CHR$(12)                                 'Page Feed
         LF$ = CHR$(10)                                 'Line Feed
         OutFile$ = "LPT1"                              'printer port
         JustCount% = 0                                 'Pause after page count off
         Tune% = 0
         PC.LineLen = 80                                'Maximum length of line

         CALL PrintLogo

         '============== Setup from the command line
         CMD$ = ucase$(command$)
         IF LEN(CMD$) THEN                               'do this only when CMD$ is used
           IF LEFT$(LTRIM$(CMD$), 1) <> "/" THEN
              IF INSTR(CMD$, "/") THEN
                 FileName$ = MID$(LTRIM$(CMD$), 1, INSTR(LTRIM$(CMD$), "/")-1 )
              ELSE
                 FileName$ = LTRIM$(CMD$)
              END IF
           END IF

           IF INSTR(CMD$, "/D") THEN
              PC.CurDate = -1                           'Do current date
              PC.DoHeader = -1
           END IF

           IF INSTR(CMD$, "/F") THEN
              PC.FileTitle = -1                         'Do file title
              PC.DoHeader = -1
           END IF

           IF INSTR(CMD$, "/P") THEN
              PC.PgNumber = -1                          'Do page numbers
              PC.DoHeader = -1
           END IF

           IF INSTR(CMD$, "/C") THEN
              JustCount% = -1                           'Just count pages
           END IF

           IF INSTR(CMD$, "/2") THEN                    'Use LPT2
              OutFile$ = "Lpt2"
           END IF

           IF INSTR(CMD$, "/W") THEN                    'Use linewrap
              PC.LineWrap = -1
           END IF

           IF INSTR(CMD$, "/S") THEN                    'Use beep statements
              Tune% = -1
           END IF


           IF INSTR(CMD$, "/X") THEN                    'Use auto duplexing
              PC.Dplx = -1
           END IF

           IF INSTR(CMD$, "/H") or INSTR(CMD$,"/?") THEN'Show help
              PRINT "Usage: PCBOOK3 filename [/F] [/P] [/D] [/C] [/2] [/A] [/W] [/S] [/H]"
              PRINT "/F - prints file name at top of page"
              PRINT "/P - prints page numbers"
              PRINT "/D - prints current date on every page"
              PRINT "/C - only reports required physical page count"
              PRINT "/2 - print to LPT2"
              PRINT "/A - prompt for alternate file to print to"
              PRINT "/W - set line wrap on"
              PRINT "/X - Print on HP compatible duplex printer"
              PRINT "     (Automatically flips pages for you)"
              PRINT "/S - sound on"
              PRINT "/H or /? provides this help message"
              GOTO OutHere
            END IF
         END IF

         '============== Open text file

GetName:
         IF LEN(FileName$) = 0 THEN
            IF Tune% THEN BEEP
            print "Enter file name to print: ";
            PDQINPUT FileName$
            PRINT
            IF FileName$ = "" THEN GOTO OutHere
         END IF                                      'Test if file is there
         if PdqExist%(Filename$)=0 then
            print "Does not exist!"
            close : end
         end if

         '============== Prompt for new output file if requested
         IF INSTR(CMD$, "/A") THEN                    'Prompt for output file
            PRINT
            IF Tune% THEN BEEP
            print "Enter alternate output file: ";
            PDQINPUT Temp$
            IF Temp$ <> "" THEN OutFile$ = Temp$      'allow a change of mind
            if OutFile$ = FileName$ then
               print "Sorry - Can't make INPUT and OUTPUT filenames the same."
               close : end
            end if
            PRINT
         END IF

         '============== Build index array for pages in FileName$
         csr% = csrlin
         CALL BuildArray(PtrArray&(), Page%)          'Built pointer array

         '============== Figure number of pages needed
         IF Page% MOD 4 THEN                          'Even multiples of 4 only
            Page% = Page% + (4 - Page% MOD 4)         '  correct for less
         END IF

         IF JustCount% THEN
            print 
            print "You will need "Page%\4;
            if page%\4 > 1 then
               print "sheets";
            else
               print "sheet";
            end if
            print " of paper."
            close : end
         END IF

         OPEN OutFile$ FOR OUTPUT AS #2               'Open printer or output file
         ercd% = err : if ercd% then goto ErrorDept
         CALL PrintSetup                              'Set up printer

        'Page parsing variables
         jj% = -1
         leftside% = Page%
         rightSide% = 1
         FirstPass% = -1

         OPEN FileName$ FOR BINARY AS #1              'Open the input file
         ercd% = err : if ercd% then goto ErrorDept

         PDQPRINT "Printing",csr%,1%,7%

         '============== Start of print routine

DoPass:
         BookMark% = (Page% \ 4)
         if Pc.Dplx then
            BookMark% = BookMark% * 2                   'Double if duplex ptr
            IF Bookmark% = 0 THEN Bookmark% = 2         'Force 2 if too small
         else
            IF Bookmark% = 0 THEN Bookmark% = 1         'Force 1 if too small
         end if

         '============== Read text and send to printer
         DO                                             'Print the right side of the page first
            pdqprint "Page: "+str$(RightSide%)+"     ",csr%,65,7
            if Pc.Dplx then
               pdqprint string$(25-((25*(Bookmark%*2))\Page%),""),csr%,15,10
            else
               pdqprint string$(25-((25*Bookmark%)\Page%),""),csr%,15,10
            end if
            IF PtrArray&(RightSide% + 1) = 0 THEN       'If blank, then skip it
               GOTO NextPage
            END IF
            PRINT #2, ESCandF$;"2y2X";                  'execute the macro

            LJLocate 95, 0                              'Home the cursor

            IF PC.DoHeader THEN CALL Header(rightSide%) 'Header if needed
            Buffer$ = SPACE$(PtrArray&(rightSide% + 1) - PtrArray&(rightSide%))

            GET #1, PtrArray&(rightSide%), Buffer$      'Read in a page

            IF INSTR(Buffer$, FF$) THEN                 'If the last character is a PF
               PRINT #2, LEFT$(Buffer$, INSTR(Buffer$, FF$) - 1); 'print only text
            ELSE
               PRINT #2, Buffer$;                       'Otherwise print full line
            END IF

NextPage:
            pdqprint "Page: "+str$(LeftSide%)+"     ",csr%,65,7
            IF PtrArray&(LeftSide% + 1) = 0 THEN        'Don't print blank pages
               GOTO NextPage1
            END IF
            PRINT #2, ESCandF$;"1y2X";          'execute the macro
            LJLocate 0, 0                               'Home the cursor
            IF PC.DoHeader THEN CALL Header(leftside%)  'Header if needed
            Buffer$ = SPACE$(PtrArray&(leftside% + 1) - PtrArray&(leftside%))                'Setup buffer for input
            IF leftside% = 0 THEN                       'If pointing at blank page, skip
               GOTO NextPage1
            END IF
            GET #1, PtrArray&(leftside%), Buffer$       'Read in a page

            IF INSTR(Buffer$, FF$) THEN                 'if the last character is a PF
               PRINT #2, LEFT$(Buffer$, INSTR(Buffer$, FF$) - 1); 'print only text
            ELSE                                        'print only text
               PRINT #2, Buffer$;                       'otherwise print all
            END IF

NextPage1:
            PRINT #2, FF$;                              'Page feed
            if Pc.dplx then
               jj% = jj% * -1                           'Toggle direction
               leftside% = leftside% - jj%              'Move
               rightSide% = rightSide% + jj%            'Move
               SWAP rightSide%, leftside%               'Swap sides
            else
               leftside% = leftside% - 2
               rightSide% = rightSide% + 2
            end if
            Bookmark% = Bookmark% - 1                   'Track our progress
         LOOP UNTIL bookmark% = 0                       'Print pages until halfway through

         pdqprint string$(25,""),csr%,15,10

         If not Pc.Dplx then
            '============== Pause between sides
            IF FirstPass THEN                           'If side one, prompt and get 2nd side
               PDQPRINT "Reinsert paper & press Enter",csr%,45,15
               IF Tune% THEN BEEP

WaitKey:                                                'Press any key to continue loop
              DO
                A% = PDQINKEY%                          'Get a key if one is pending
              LOOP UNTIL a%                             'Integer compares faster than strings
              IF A%  = 27 THEN GOTO PrtReset            'ESC key, takes you out
              IF A% <> 13 THEN GOTO WaitKey             'Enter key only to prevent accidentally
                                                        ' starting printer
              PDQPRINT space$(28),csr%,45,10
              FirstPass = 0                             'Flag for second pass
              GOTO DoPass
            END IF                                      'End of first pass

         end if
         PRINT
         PRINT "Printing completed"
         IF Tune% THEN BEEP

PrtReset:
         PRINT #2, ESC$; "E";                           'Reset laserjet

OutHere:
         CLOSE                                          'Close all files
         END                                            'Thats all for now

        '============== Error handler
ErrorDept:
         PRINT
         PRINT "*** Error ***"
         BEEP
         SELECT CASE ERCD%
            CASE 24
               PRINT "Device timed out"
            CASE 25                                   'Device fault
               PRINT "Device Fault"
            CASE 27                                   'Paper is out
               PRINT "Out of paper"
            CASE 53                                   'Source file not there
               PRINT "File "; FileName$; " not found"
               FileName$ = ""
               GOSUB AWayOut
               goto GetName
            CASE 71                                   'Open drive door
               PRINT "Disk drive not ready"
            CASE ELSE
               PRINT "Error number "; ERCD%
         END SELECT
         gosub AwayOut
         close : end

AWayOut:
         PRINT
         PRINT "Press any key to exit."

        '============================ End of main module ============================

        SUB BuildArray (PtrArray&(), Pgcount%) STATIC
           'FileName$ is shared from the main module

           print

           MaxLines% = 66                               'Maximum number of lines
           Offset& = 1                                  'Start of file (seek point)
           OPEN FileName$ FOR BINARY AS #1 LEN = 1      'Open file to check
           TotalSize& = LOF(1)                          'Get LEN of file so we don't read too far
           print "              0%   25%   50%   75%   100%"
           print "              Ŀ
           csr% = csrlin
           PDQPRINT "Analyzing",csr%,1%,7%
           pdqprint string$(25,""),csr%,15%,6%         'Print bar graph
           FileLeft& = TotalSize&                       'Setup a counter to show whats left
           MemAvail& = FRE(FileName$) - 2048            'Check available string memory
           IF MemAvail& < 2048 THEN ERROR 14            'Force out of memory error
           SixteenK% = 16384

           IF TotalSize& > SixteenK% THEN               'Set a buffer size
              IF MemAvail& > SixteenK% THEN             'If the file is larger than 16K
                 BufSize& = SixteenK%                   'Set it to 16k
              ELSE
                 BufSize% = MemAvail&
              END IF
           ELSE
              IF TotalSize& < MemAvail& THEN            'Otherwise set it to file size
                 BufSize& = TotalSize&
              END IF
           END IF

           Pgcount% = 1                                 'Initialize page count
           PtrArray&(Pgcount%) = 1                      'First pointer is always 1
           LnCount% = 0                                 'Initialize line count

GetPage:
                                                        'Read the file
          IF FileLeft& < BufSize& THEN                  'Check amount left to read
             Buffer$ = SPACE$(FileLeft&)                'If less than our buffer, use lessor
          ELSE
             Buffer$ = SPACE$(BufSize&)                 'Otherwise use full buffer size
          END IF

          GET #1, Offset&, Buffer$                      'Read in a buffers worth

          pdqprint string$((25&*offset&)\TotalSize&,""),csr%,15%,14%

          StPtr% = 1                                    'Pointer into buffer$
          LastLine% = 0                                 'remember last position

PageCheck:
          TempLn% = INSTR(StPtr%, Buffer$, LF$)         'Position of next linefeed
          TempPg% = INSTR(StPtr%, Buffer$, FF$)         'Position of next pagefeeds

          IF TempPg% THEN                               'If there was a page feed
             IF TempPg% < TempLn% OR TempLn% = 0 THEN   '  was it before our linefeed?
                Pgcount% = Pgcount% + 1                 '  yes then bump page count
                PtrArray&(Pgcount%) = Offset& + TempPg% '  set next array element
                StPtr% = TempPg% + 1                    '  set instr pointer
                LnCount% = 0                            '  reset linecount
                IF StPtr% < LEN(Buffer$) THEN GOTO PageCheck 'and loop back for more
              END IF
          END IF

          IF TempLn% THEN                               'Linefeed
             IF PC.LineWrap THEN                        'If /W the check line length
                IF TempLn% - StPtr% > PC.LineLen THEN   'Greater than 80?
                   DO                                   'check for line wrap
                     LnCount% = LnCount% + 1            'increment line
                     IF LnCount% = MaxLines THEN GOTO PageBreak  '> 66 lines
                     StPtr% = StPtr% + PC.LineLen
                   LOOP WHILE TempLn% - StPtr% > PC.LineLen
                END IF
             END IF
             LnCount% = LnCount% + 1                    'Increment page count

PageBreak:
             IF LnCount% = MaxLines% THEN
                 Pgcount% = Pgcount% + 1
                    IF Pgcount% > 512 THEN
                       PRINT "Too many pages- printing only 512"
                       GOTO EndBuild
                    END IF
                 PtrArray&(Pgcount%) = Offset& + TempLn% 'point to next in point in file
                 LnCount% = 0
             END IF
             StPtr% = TempLn% + 1                       'point ahead 1 byte for next scan

             IF StPtr% <= LEN(Buffer$) THEN
                GOTO PageCheck                          'keep checking
             END IF
          END IF

          Offset& = Offset& + LEN(Buffer$)              'Pointer into file (tally)
          StPtr% = 1                                    'Reset Buffer pointer
          FileLeft& = TotalSize& - Offset&              'Calculate how much is left
          IF Offset& < TotalSize& THEN GOTO GetPage     'If more text in file, keep going
         
EndBuild:
          PtrArray&(Pgcount% + 1) = TotalSize&          'Set last pointer to end of file

          CLOSE #1                                      'Close input file
          pdqprint string$(25,""),csr%,15%,14%
          print

         END SUB                                        'End of BuildArray Module

         SUB EndMacro (Num$) STATIC
            PRINT #2, ESCandF$; Num$; "y1X";            'Send end of macro command
            PRINT #2, ESCandF$; Num$; "y9X";            'Make it temporary (10 to be permanent)
         END SUB

         SUB Header (Page%) STATIC
           Hdr$ = SPACE$(PC.LineLen)                    'Create a string to print
           IF PC.FileTitle THEN                         'Print the filename
              MID$(Hdr$, 40 - LEN(FileName$) \ 2) = UCASE$(FileName$)
           END IF

           IF PC.PgNumber THEN                          'Print the current page
             PTemp$ = "Page " + STR$(Page%)
             IF Page% MOD 2 THEN
                MID$(Hdr$, PC.LineLen - LEN(PTemp$)) = PTemp$ 'odd page, right side
             ELSE
                MID$(Hdr$, 1) = PTemp$                  'even page, left side
             END IF
           END IF

           IF PC.CurDate THEN                           'Print the current date
             IF Page% MOD 2 THEN
                MID$(Hdr$, 1) = DATE$                   'even page, left side
             ELSE
                 MID$(Hdr$, PC.LineLen - LEN(DATE$)) = DATE$ 'odd page, right side
             END IF
           END IF
           PRINT #2, Hdr$                               'Print the Header
           PRINT #2,                                    ' and skip a line for readability

         END SUB
         
         SUB LJLocate (X%, Y%) STATIC                   'Laser Jet cursor locate
            Temp$ = ESC$ + "&a" + STR$(Y%) + "r" + STR$(X%) + "C"
            PRINT #2, Temp$;
         END SUB

         SUB PrintLogo STATIC                           'Banner logo
           locate 25,1
           PRINT
           PRINT STRING$(79,"")
           PRINT "PCBook3 - PC Magazine Booklet Printing Utility"
           PRINT "Copyright 1990 PC Magazine  Ziff Davis  Jay Munro"
           print "Modifications for HP duplexing lasers: Mike Shaffer, Savant Technologies, Inc."
           PRINT STRING$(79,"")
         END SUB

         SUB PrintSetup                                 '============== Send codes to prepare printer
            PRINT #2, ESC$; "E";                        'Reset laserjet (simple isn't it!)
            PRINT #2, ESC$; "&l1o5.45C";                'Select lineprinter font"
            PRINT #2, ESC$; "(s0p16.66H";               '  and pitch
            PRINT #2, ESC$; "&l0L";                     'Turn off page feed at 66 lines
            if PC.Dplx then
               PRINT #2, ESC$; "&l2S";                  'Turn on duplexing
            end if

            IF PC.LineWrap THEN                         'Wrap lines > 80 chars
               PRINT #2, ESC$; "&s0C";
            END IF

            PRINT #2, ESC$; "&l2E";                     'Top margin 2 lines

            CALL StartMacro("1")                        'Left side macro
                 PRINT #2, ESC$; "9";                   'Reset left - right margins
                 PRINT #2, ESC$; "&a0l80M";             'set left margin 0, right 80
            CALL EndMacro("1")

            CALL StartMacro("2")                        'Right side macro
                 PRINT #2, ESC$; "9";                   'Reset left - right margins
                 PRINT #2, ESC$; "&a95l175M";           'set left margin 95, right 175
            CALL EndMacro("2")

         END SUB

         SUB StartMacro (Num$) STATIC
            PRINT #2, ESCandF$; Num$; "Y";              'Macro will have an id of Num$
            PRINT #2, ESCandF$;"0X";                    'Start the macro now
         END SUB

